
#include	<Box2D/Box2D.h>
#include	<time.h>
#include	<random>

#define DLLAPI extern "C" __declspec(dllexport)
#include	"CurlingSimulator.h"

#ifdef _DEBUG
	#pragma comment( lib, "Box2D_d.lib" )
#endif
#ifndef _DEBUG
	#pragma comment( lib, "Box2D.lib" )
#endif

// Xg[
#define STONEINFO_SIZE 0.145f
#define STONEINFO_ANGLE 0.0f
#define STONEINFO_DYNAMIC true
#define STONEINFO_ISBALL true
#define STONEINFO_DENSITY 10.0f
#define STONEINFO_RESTITUTION 1.0f
#define STONEINFO_FRICTION 12.009216f
//#define STONEINFO_FRICTION 0.12009216f
#define STONEINFO_CATEGORYBITS 3
#define STONEINFO_MASKBITS 3

#define RINKINFO_X 0.0f
#define RINKINFO_Y 3.05f
#define RINKINFO_W 4.75f
#define RINKINFO_H 42.5f

#define TEEPOSITION_X 2.375f
#define TEEPOSITION_Y 4.88f

#define PLAYAREA_X 0.0f
#define PLAYAREA_Y 3.05f
#define PLAYAREA_W 4.75f
#define PLAYAREA_H 8.23f

enum {
	STATE_RINK			= 0x0001,					// N
	STATE_PLAYAREA		= STATE_RINK << 1,			// vCGA
	STATE_FREEGUARD		= STATE_PLAYAREA << 1,		// t[K[h][
	STATE_HOUSE			= STATE_FREEGUARD << 1		// nEX
};

// n`\
typedef struct _GROUND{
	b2Body *body;
} GROUND, *LPGROUND;

// Q[
typedef struct _SUBGAMESTATE{
	int		ShotNum;
	b2Body	*body[16];

} SUBGAMESTATE, *PSUBGAMESTATE;

// Xg[
typedef struct _StoneInfo{
	float	RangeWithError_x;
	float	RangeWithError_y;

} STONEINFO, *PSTONEINFO;

// \TCY
float	VIWSC = 32.0f;

b2Vec2 gravity( 0.0f, 0.0f );
bool doSleep = true;

// FPS
float g_timeStep = 1.0f / 600.0f;
// V~[Vx
int velocityIterations = 10;
int positionIterations = 10;

// NEv[GA
STONEINFO StoneInfo;

// ]x
#define STANDARD_ANGLE 0.066696f

// sNZ  [g
float ptom( int x )
{
	return	(float)( x/VIWSC );
}

// [g  sNZ
int mtop( float x )
{
	return	(int)( x*VIWSC );
}

// {fB[̍쐬
//=====================================================================================
b2Body *CreateBody( float x, float y, float w, 
	float h, float angle, bool dynamic, int isBall, 
	float density, float restitution, float friction, int categoryBits, int maskBits, b2World *world )
{
	b2BodyDef bodyDef;

	// I^ÓI
	if( dynamic ){
		bodyDef.type = b2_dynamicBody;
	}
	else{
		bodyDef.type = b2_staticBody;
	}

	// \ʒu
	bodyDef.position.Set( x, y );

	// px
	bodyDef.angle = angle;

	// {fB̐
	b2Body *body = world->CreateBody( &bodyDef );


	// tBNX`̍쐬
	//-------------------------------------------------------
	b2FixtureDef fixtureDef;

	b2CircleShape dynamicBall;
	dynamicBall.m_radius = w;

	b2PolygonShape staticBox;
	staticBox.SetAsBox( w, h );

	b2PolygonShape staticTri;
	staticTri.SetAsBox( w, h );
	staticTri.m_vertexCount = 3;
	staticTri.m_vertices[0].Set( 0, 0 );
	staticTri.m_vertices[1].Set( w, 0 );
	staticTri.m_vertices[2].Set( w/2, h*sqrt( 3.0f )/2 );


	if( isBall == 1 ){
		fixtureDef.shape = &dynamicBall;
	}
	else if( isBall == 0 ){
		fixtureDef.shape = &staticBox;
	}
	else{
		fixtureDef.shape = &staticTri;
	}

	// x
	fixtureDef.density = density;

	// 
	fixtureDef.restitution = restitution;

	// C
	fixtureDef.friction = friction;

	fixtureDef.filter.categoryBits = categoryBits;
	fixtureDef.filter.maskBits = maskBits;

	body->CreateFixture( &fixtureDef );

	return body;
}

bool SetShot( SHOTVEC Shot, SUBGAMESTATE *pSubGameState, b2World *world, float RandX, float RandY, SHOTVEC *lpResShot )
{
	bool bRet = false;
	if( pSubGameState->ShotNum < 16 ){
		pSubGameState->body[ pSubGameState->ShotNum ] = CreateBody( RINKINFO_W/2, 41.28f, STONEINFO_SIZE, STONEINFO_SIZE, STONEINFO_ANGLE, STONEINFO_DYNAMIC, STONEINFO_ISBALL, STONEINFO_DENSITY, STONEINFO_RESTITUTION, STONEINFO_FRICTION, STONEINFO_CATEGORYBITS, STONEINFO_MASKBITS, world );

		lpResShot->x = Shot.x + RandX;
		lpResShot->y = Shot.y + RandY;
		lpResShot->angle = Shot.angle;

		pSubGameState->body[ pSubGameState->ShotNum ]->SetLinearVelocity( b2Vec2( lpResShot->x, lpResShot->y ) );
		if( lpResShot->angle == 0 ){
			pSubGameState->body[ pSubGameState->ShotNum ]->SetAngularVelocity( STANDARD_ANGLE );
		}
		else{
			pSubGameState->body[ pSubGameState->ShotNum ]->SetAngularVelocity( STANDARD_ANGLE * -1 );
		}
		pSubGameState->ShotNum++;

		bRet = true;
	}

	return bRet;
}

// C͂̌vZ
//=====================================================================================
b2Vec2 FrictionStep( float friction, b2Vec2 vec, float angle )
{
	b2Vec2	velocity = vec;
	float length = velocity.Length();

	if( length > friction ){
		b2Vec2 normalize, normalize2;
		normalize = velocity;

		normalize.Normalize();
		normalize2.y = normalize.x;
		normalize2.x = -1*normalize.y;

		normalize *= friction;
		velocity -= normalize;

		// ]ɂe
		if( angle != 0.0f ){
			normalize2 *=( ( angle > 0 )? STANDARD_ANGLE * friction: -STANDARD_ANGLE * friction );
			length = velocity.Length();
			velocity += normalize2;
			velocity.Normalize();
			velocity *= length;
		}

		return velocity;
	}
	else{
		return b2Vec2( 0.0f, 0.0f );
	}
}

void Friction( float friction, SUBGAMESTATE *pSubGameState )
{
	b2Vec2 vec;

	for( int i=0; i<pSubGameState->ShotNum; i++ ){
		if( pSubGameState->body[i] != nullptr ){
			vec = FrictionStep( friction, pSubGameState->body[i]->GetLinearVelocity(), pSubGameState->body[i]->GetAngularVelocity() );
			pSubGameState->body[i]->SetLinearVelocity( vec );
			if( vec.Length() == 0 ){
				pSubGameState->body[i]->SetAngularVelocity( 0.0f );
			}
		}
	}
}

// Xg[̈ʒu
int GetStoneState( float x, float y )
{
	int iRet = 0;
	b2Vec2 WorkVec = b2Vec2( x - 2.375f, y - 4.88f );

	// N
	if( ( x > RINKINFO_X + STONEINFO_SIZE && x < RINKINFO_X + RINKINFO_W - STONEINFO_SIZE )
		&& ( y > RINKINFO_Y + STONEINFO_SIZE && y < RINKINFO_Y + RINKINFO_H - STONEINFO_SIZE ) )
	{
		iRet |= STATE_RINK;
	}

	// vCGA
	if( ( x > PLAYAREA_X + STONEINFO_SIZE && x < PLAYAREA_X + PLAYAREA_W - STONEINFO_SIZE )
		&& ( y > PLAYAREA_Y - STONEINFO_SIZE && y < PLAYAREA_Y + PLAYAREA_H - STONEINFO_SIZE ) )
	{
		iRet |= STATE_PLAYAREA;
	}

	// nEX
	if( WorkVec.Length() < 1.83f + STONEINFO_SIZE ){
		iRet |= STATE_HOUSE;
	}

	// t[K[h][
	if( !( iRet & STATE_HOUSE )
		&& ( x > PLAYAREA_X + STONEINFO_SIZE && x < PLAYAREA_X + PLAYAREA_W - STONEINFO_SIZE )
		&& ( y > 4.88 - STONEINFO_SIZE && y < PLAYAREA_Y + PLAYAREA_H - STONEINFO_SIZE ) )
	{
		iRet |= STATE_FREEGUARD;
	}

	return iRet;
}

// V~[V - C[v
//=====================================================================================
int MainLoop( SUBGAMESTATE *pSubGameState, b2World *world, float timeStep, float *pLoci, size_t ResLociSize, int LoopCount = -1 )
{
	int iRet;
	b2Vec2 vec;

	Friction( STONEINFO_FRICTION*timeStep*0.5f, pSubGameState );
	for( iRet=0; iRet < LoopCount || LoopCount == -1; iRet++ ){
		// C͂̌vZ
		world->Step( timeStep, velocityIterations, positionIterations );
		Friction( STONEINFO_FRICTION*timeStep, pSubGameState );

		// Vbg̋OՂzɊi[
		if(pLoci != nullptr){
			for( int i=0; i<16 && ((iRet*32 + ((i+i) * 2)) * sizeof(float) < ResLociSize); i++ ){
				if(i < pSubGameState->ShotNum && pSubGameState->body[i] != nullptr){
					vec = pSubGameState->body[i]->GetPosition();
					pLoci[iRet*32 + i * 2] = vec.x;
					pLoci[iRet*32 + i * 2 + 1] = vec.y;
				}
				else{
					pLoci[iRet*32 + i * 2] = 0.0f;
					pLoci[iRet*32 + i * 2 + 1] = 0.0f;
				}
			}
		}


		/*
			SẴXg[Î~ true Ԃ
			Xg[Noꍇɂ͂̃Xg[O
		*/
		for( int i=0; i<pSubGameState->ShotNum; i++ ){
			if( pSubGameState->body[i] != nullptr ){
				// Xg[NOɏo珜O
				vec = pSubGameState->body[i]->GetPosition();
				if( GetStoneState( vec.x, vec.y ) == 0 ){
					world->DestroyBody( pSubGameState->body[i] );
					pSubGameState->body[i] = nullptr;
				}
				else if( pSubGameState->body[i]->IsAwake() == true ){
					break;
				}
			}

			if( i == pSubGameState->ShotNum - 1 ){
				if( LoopCount > iRet ){
					iRet = -1;
				}
				goto LOOP_END;
			}
		}
	}
LOOP_END:

	return iRet;
}

// _vZ
//=====================================================================================
DLLAPI int GetScore( GAMESTATE *pGameState )
{
	/*
		葤̓_ɃXRAԂB
		肪_ꍇɂ͐̐A肪_ꍇɂ͕̐ԂB
	*/

	int iScore = 0;
	float NearestLength_r, NearestLength_y, House;
	b2Vec2 WorkVec, TeePos;

	TeePos.Set( 2.375f, 4.88f );
	NearestLength_r = NearestLength_y = House = ( 1.83f + 0.15f );

	// No.1Xg[
	for( int i=0; i<=pGameState->ShotNum; i+=2 ){
		if( pGameState->body[i] != nullptr ){
			WorkVec = b2Vec2( pGameState->body[i][0], pGameState->body[i][1] );
			WorkVec.operator -=( TeePos );
			if( WorkVec.Length() <= NearestLength_r ){
				NearestLength_r = WorkVec.Length();
			}
		}
	}

	// No.1Xg[
	for( int i=1; i<=pGameState->ShotNum; i+=2 ){
		if( pGameState->body[i] != nullptr ){
			WorkVec = b2Vec2( pGameState->body[i][0], pGameState->body[i][1] );
			WorkVec.operator -=( TeePos );
			if( WorkVec.Length() <= NearestLength_y ){
				NearestLength_y = WorkVec.Length();
			}
		}
	}

	// _̌vZ
	if( NearestLength_r < NearestLength_y ){
		for( int i=0; i<=pGameState->ShotNum; i+=2 ){
			if( pGameState->body[i] != nullptr ){
				WorkVec = b2Vec2( pGameState->body[i][0], pGameState->body[i][1] );
				WorkVec.operator -=( TeePos );
				if( WorkVec.Length() < NearestLength_y ){
					iScore++;
				}
			}
		}
	}
	else if( NearestLength_y < NearestLength_r ){
		for( int i=1; i<=pGameState->ShotNum; i+=2 ){
			if( pGameState->body[i] != nullptr ){
				WorkVec = b2Vec2( pGameState->body[i][0], pGameState->body[i][1] );
				WorkVec.operator -=( TeePos );
				if( WorkVec.Length() < NearestLength_r ){
					iScore--;
				}
			}
		}
	}

	if( ( pGameState->ShotNum % 2 == 0 && pGameState->WhiteToMove != 0 )
		|| ( pGameState->ShotNum % 2 == 1 && pGameState->WhiteToMove == 0 ) )
	{
		iScore *= -1;
	}

	return iScore;
}

// Vbg̐
//=====================================================================================
b2Vec2 CreateShot( float x, float y )			/* Vbg̋𖀎C͂llɒ */
{
	b2Vec2 Shot;
	float len;

	Shot.Set( x-2.375f, y-41.28f );
	len = Shot.Length();
	Shot.Normalize();
	Shot.operator *=( sqrt( len*2.0f*STONEINFO_FRICTION ) );

	return Shot;
}

// Vbg̒~ʒu̍WVbg𐶐
//-------------------------------------------------------
DLLAPI int CreateShot( SHOTPOS Shot, SHOTVEC *lpResShot )
{
	float tt = 0.0335f;
	b2Vec2 vec, TeePos;
	TeePos.Set( 2.375f, 4.88f );

	if( Shot.angle == true ){
		vec = CreateShot( 1.22f + Shot.x - tt*( Shot.y - TeePos.y ), tt*( Shot.x - TeePos.x ) + Shot.y );
	}
	else{
		vec = CreateShot( -1.22f + Shot.x + tt*( Shot.y - TeePos.y ), -tt*( Shot.x - TeePos.x ) + Shot.y );
	}

	lpResShot->x = vec.x;
	lpResShot->y = vec.y;
	lpResShot->angle = Shot.angle;

	return true;
}

// Vbg̕EE]Vbg𐶐
//-------------------------------------------------------
bool CreateShot( SHOTPOS Shot, float Power, SHOTVEC *lpResShot )
{
	bool bRet = false;
	SHOTPOS tmpShot;
	float StY;

	if( Power >= 0.0f ){
		if( Power >= 0.0f && Power <= 1.0f ){
			StY = 11.28f + ( ( 0.0f - Power ) * 0.76f );
		}
		else if( Power > 1.0f && Power <= 3.0f ){
			StY = 10.52f + ( ( 1.0f - Power ) * 1.52f );
		}
		else if( Power > 3.0f && Power <= 4.0f ){
			StY = 7.47f + (  ( 3.0f - Power ) * 0.76f );
		}
		else if( Power > 4.0f && Power <= 10.0f ){
			StY = 6.71f + ( ( 4.0f - Power ) * 0.61f );
		}
		else if( Power > 10.0f ){
			if( Power > 16 ){
				Power = 16;
			}
			StY = 3.05f + ( ( 10.0f - Power ) * 1.52f );
		}

		tmpShot.x = ( ( Shot.x - 2.375f ) * ( StY - 41.28f ) / ( Shot.y - 41.28f ) ) + 2.375f;
		tmpShot.y = StY;
		tmpShot.angle = Shot.angle;

		CreateShot( tmpShot, lpResShot );
		bRet = true;
	}
	else{
		CreateShot( Shot, lpResShot );
		bRet = true;
	}

	return bRet;
}

// b2Body ̔j
//=====================================================================================
void DestroyBody( SUBGAMESTATE *pSubGameState, b2World *world )
{
	for( int i=0; i<pSubGameState->ShotNum; i++ ){
		if( pSubGameState->body[i] != nullptr ){
			world->DestroyBody( pSubGameState->body[i] );
			pSubGameState->body[i] = nullptr;
		}
	}
}

// Q[̏
//-------------------------------------------------------
bool InitSUBGAMESTATE( SUBGAMESTATE *pSubGameState, GAMESTATE *pGameState, b2World *world )
{
	memset( pSubGameState, 0x00, sizeof( SUBGAMESTATE ) );
	pSubGameState->ShotNum = pGameState->ShotNum;

	if( pGameState->ShotNum > 15 ){
		return false;
	}

	for( int i=0; i<pGameState->ShotNum; i++ ){
		if(  GetStoneState( pGameState->body[i][0], pGameState->body[i][1] ) & STATE_PLAYAREA ){
			pSubGameState->body[i] = CreateBody( pGameState->body[i][0], pGameState->body[i][1], STONEINFO_SIZE, STONEINFO_SIZE, STONEINFO_ANGLE, STONEINFO_DYNAMIC, STONEINFO_ISBALL,  STONEINFO_DENSITY, STONEINFO_RESTITUTION, STONEINFO_FRICTION, STONEINFO_CATEGORYBITS, STONEINFO_MASKBITS, world );
		}
	}

	return true;
}


// ʉ߈ʒuEE]qbgVbg𐶐
//-------------------------------------------------------
DLLAPI int CreateHitShot( SHOTPOS Shot, float Power, SHOTVEC *lpResShot )
{
	int iRet = 0;

	// SUBGAMESTATȄ
	SUBGAMESTATE SubGameState;
	b2World world( gravity, doSleep );
	memset( &SubGameState, 0x00, sizeof( SUBGAMESTATE ) );
	SubGameState.ShotNum = 0;

	// Vbg̃Zbg
	SHOTVEC tShot;
	CreateShot( Shot, Power, &tShot );

	SetShot( tShot, &SubGameState, &world, 0.0f, 0.0f, lpResShot );

	b2Vec2 tar, s, t;
	tar.Set( Shot.x - 2.375f, 41.28f - Shot.y );
	Friction( STONEINFO_FRICTION*g_timeStep*0.5f, &SubGameState );
	for(int i=0; ; i++){

		// C͂̌vZ
		world.Step( g_timeStep, velocityIterations, positionIterations );
		Friction( STONEINFO_FRICTION*g_timeStep, &SubGameState );

		t = s;
		s = SubGameState.body[SubGameState.ShotNum-1]->GetPosition();
		s = b2Vec2( s.x - 2.375f, 41.28f - s.y );
		if( s.Length() > tar.Length() ){
			b2Vec2 u = s - t;
			float a = s.Length() - t.Length();
			float b = tar.Length() - t.Length();

			u *= b/a;
			t += u;

			t.Length();
			float n = atan2( tar.y, tar.x );
			float m = atan2( t.y, t.x );

			n = m - n;

			u = b2Vec2( tShot.x, tShot.y );
			lpResShot->x = cos( n )*u.x - sin( n )*u.y;
			lpResShot->y = sin( n )*u.x + cos( n )*u.y;
			lpResShot->angle = Shot.angle;

			iRet = 1;
			goto EndOfCreateHitShot;
		}

		// SẴXg[Î~ true Ԃ
		for( int i=0; i<SubGameState.ShotNum; i++ ){
			if( SubGameState.body[i] != nullptr && SubGameState.body[i]->IsAwake() == true ){
				break;
			}

			if( i == SubGameState.ShotNum - 1 ){
				goto EndOfCreateHitShot;
			}
		}
	}

EndOfCreateHitShot:

	return iRet;
}

// V~[V{
//=====================================================================================
DLLAPI int Simulation( GAMESTATE *pGameState, SHOTVEC Shot, float Rand = STONEINFO_SIZE, SHOTVEC *lpResShot = nullptr, int LoopCount = -1 )
{
	SUBGAMESTATE SubGameState;
	b2Vec2 vec;
	bool bFreeGuardRule;
	SHOTVEC WorkVec;

	// K
	std::random_device seed_gen;
	std::default_random_engine engine(seed_gen());

	// Q[̏
	b2World world( gravity, doSleep );
	if( InitSUBGAMESTATE( &SubGameState, pGameState, &world ) == false ){
		return false;
	}

	//  0, W΍Xg[aŕz
	if(Rand != 0.0f){
		std::normal_distribution<float> dist(0, Rand);

		SHOTPOS TeePos, ShotPos;
		SHOTVEC TeeShot, ShotVec;

		TeePos.x = 2.375f;
		TeePos.y = 4.88f;
		CreateShot(TeePos, &TeeShot);

		ShotPos.x = 2.375f + dist(engine);
		ShotPos.y = 4.88f + dist(engine);
		CreateShot(ShotPos, &ShotVec);
		
		// Vbg̃Zbg
		if( SetShot( Shot, &SubGameState, &world, TeeShot.x - ShotVec.x, TeeShot.y - ShotVec.y, &WorkVec ) == false ){
			return false;
		}
	}
	else{
		// Vbg̃Zbg
		if( SetShot( Shot, &SubGameState, &world, 0.0f, 0.0f, &WorkVec ) == false ){
			return false;
		}
	}

	// 
	if( lpResShot != nullptr ){
		memcpy( lpResShot, &WorkVec, sizeof( SHOTVEC ) );
	}

	// Z
	if( MainLoop( &SubGameState, &world, g_timeStep, nullptr, 0x00,  LoopCount ) == -1 ){
		return false;
	}

	// t[K[h][̓Kp
	if( SubGameState.ShotNum < 5 ){
		for( int i=(pGameState->ShotNum+1)%2; i<pGameState->ShotNum; i+=2 ){
			if( GetStoneState( pGameState->body[i][0], pGameState->body[i][1] ) & STATE_FREEGUARD ){
				if( SubGameState.body[i] == nullptr ){
					bFreeGuardRule = true;
					break;
				}
				else{
					vec = SubGameState.body[i]->GetPosition();
					if( !( GetStoneState( vec.x, vec.y ) & STATE_PLAYAREA ) ){
						bFreeGuardRule = true;
						break;
					}
				}
			}
		}
	}

	// V~[Vʂ߂
	if( bFreeGuardRule == true ){
		pGameState->body[pGameState->ShotNum][0] = 0.0f;
		pGameState->body[pGameState->ShotNum][1] = 0.0f;
	}
	else{
		for( int i=0; i<16; i++ ){
			if( SubGameState.body[i] != nullptr ){
				vec = SubGameState.body[i]->GetPosition();
			}
			else{
				vec = b2Vec2( 0.0f, 0.0f );
			}
			pGameState->body[i][0] = vec.x;
			pGameState->body[i][1] = vec.y;
		}
	}

	// VbgƎԂ̍XV
	if( SubGameState.ShotNum == 16 ){
		pGameState->Score[pGameState->CurEnd] = GetScore( pGameState );

		if( pGameState->Score[pGameState->CurEnd] > 0 ){
			pGameState->WhiteToMove = 0;
		}
		else if( pGameState->Score[pGameState->CurEnd] < 0 ){
			pGameState->WhiteToMove = 1;
		}
		else{
			pGameState->WhiteToMove ^= 1;
		}

		pGameState->ShotNum = 0;
		pGameState->CurEnd++;
	}
	else{
		pGameState->ShotNum = SubGameState.ShotNum;
		pGameState->WhiteToMove ^= 1;
		pGameState->Score[pGameState->CurEnd] = GetScore( pGameState );
	}

	// I
	DestroyBody( &SubGameState, &world );

	return true;
}

// V~[V{(g)
//=====================================================================================
DLLAPI int SimulationEx( GAMESTATE *pGameState, SHOTVEC Shot, float RandX, float RandY, SHOTVEC *lpResShot, float *pLoci, int ResLociSize )
{
	int iRet = 0;
	SUBGAMESTATE SubGameState;
	b2Vec2 vec;
	bool bFreeGuardRule;
	SHOTVEC WorkVec;

	// K
	std::random_device seed_gen;
	std::default_random_engine engine(seed_gen());

	// Q[̏
	b2World world( gravity, doSleep );
	if( InitSUBGAMESTATE( &SubGameState, pGameState, &world ) == false ){
		return false;
	}

	//  0, W΍Xg[aŕz
	SHOTPOS TeePos, ShotPos;
	SHOTVEC TeeShot, ShotVec;

	TeePos.x = ShotPos.x = 2.375f;
	TeePos.y = ShotPos.y = 4.88f;
	CreateShot(TeePos, &TeeShot);

	if(RandX != 0.0f){
		std::normal_distribution<float> distX(0, RandX);
		ShotPos.x += distX(engine);
	}
	if(RandY != 0.0f){
		std::normal_distribution<float> distY(0, RandY);
		ShotPos.y += 4.88f + distY(engine);
	}
	CreateShot(ShotPos, &ShotVec);
		
	// Vbg̃Zbg
	if( SetShot( Shot, &SubGameState, &world, TeeShot.x - ShotVec.x, TeeShot.y - ShotVec.y, &WorkVec ) == false ){
		return false;
	}

	// 
	if( lpResShot != nullptr ){
		memcpy( lpResShot, &WorkVec, sizeof( SHOTVEC ) );
	}

	// Z
	if( (iRet = MainLoop( &SubGameState, &world, g_timeStep, pLoci, ResLociSize, -1 )) == -1 ){
		return false;
	}

	// t[K[h][̓Kp
	if( SubGameState.ShotNum < 5 ){
		for( int i=( pGameState->WhiteToMove^1 ); i<pGameState->ShotNum; i+=2 ){
			if( GetStoneState( pGameState->body[i][0], pGameState->body[i][1] ) & STATE_FREEGUARD ){
				if( SubGameState.body[i] == nullptr ){
					bFreeGuardRule = true;
					break;
				}
				else{
					vec = SubGameState.body[i]->GetPosition();
					if( !( GetStoneState( vec.x, vec.y ) & STATE_PLAYAREA ) ){
						bFreeGuardRule = true;
						break;
					}
				}
			}
		}
	}

	// V~[Vʂ߂
	if( bFreeGuardRule == true ){
		pGameState->body[pGameState->ShotNum][0] = 0.0f;
		pGameState->body[pGameState->ShotNum][1] = 0.0f;
	}
	else{
		for( int i=0; i<16; i++ ){
			if( SubGameState.body[i] != nullptr ){
				vec = SubGameState.body[i]->GetPosition();
			}
			else{
				vec = b2Vec2( 0.0f, 0.0f );
			}
			pGameState->body[i][0] = vec.x;
			pGameState->body[i][1] = vec.y;
		}
	}

	// VbgƎԂ̍XV
	if( SubGameState.ShotNum == 16 ){
		pGameState->Score[pGameState->CurEnd] = GetScore( pGameState );

		if( pGameState->Score[pGameState->CurEnd] > 0 ){
			pGameState->WhiteToMove = 0;
		}
		else if( pGameState->Score[pGameState->CurEnd] < 0 ){
			pGameState->WhiteToMove = 1;
		}
		else{
			pGameState->WhiteToMove ^= 1;
		}

		pGameState->ShotNum = 0;
		pGameState->CurEnd++;
	}
	else{
		pGameState->ShotNum = SubGameState.ShotNum;
		pGameState->WhiteToMove ^= 1;
		pGameState->Score[pGameState->CurEnd] = GetScore( pGameState );
	}

	// I
	DestroyBody( &SubGameState, &world );

	return iRet;
}